@@ -30,7 +30,6 @@
 
 #include "xf86.h"
 #include "xf86_OSproc.h"
-#include "xf86_ansic.h"
 #include "xf86Priv.h"
 
 #include "xaalocal.h"
@@ -339,7 +338,9 @@ static void SAVAGEWakeupHandler( int screenNum, pointer wakeupData,
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    SavagePtr psav = SAVPTR(pScrn);
 
-   DRILock(pScreen, 0);
+   psav->pDRIInfo->wrap.WakeupHandler = psav->coreWakeupHandler;
+   (*psav->pDRIInfo->wrap.WakeupHandler) (screenNum, wakeupData, result, pReadmask);
+   psav->pDRIInfo->wrap.WakeupHandler = SAVAGEWakeupHandler;
    psav->LockHeld = 1;
    if (psav->ShadowStatus) {
       /* fetch the global shadow counter */
@@ -351,7 +352,10 @@ static void SAVAGEWakeupHandler( int screenNum, pointer wakeupData,
 #endif
       psav->ShadowCounter = psav->ShadowVirtual[1023] & 0xffff;
    }
-   psav->AccelInfoRec->NeedToSync = TRUE;
+   if (psav->useEXA)
+	exaMarkSync(pScreen);
+   else
+	psav->AccelInfoRec->NeedToSync = TRUE;
    /* FK: this flag doesn't seem to be used. */
 }
 
@@ -377,7 +381,9 @@ static void SAVAGEBlockHandler( int screenNum, pointer blockData,
       psav->ShadowVirtual[1023] = globalShadowCounter;
    }
    psav->LockHeld = 0;
-   DRIUnlock(pScreen);
+   psav->pDRIInfo->wrap.BlockHandler = psav->coreBlockHandler;
+   (*psav->pDRIInfo->wrap.BlockHandler) (screenNum, blockData, pTimeout, pReadmask);
+   psav->pDRIInfo->wrap.BlockHandler = SAVAGEBlockHandler;
 }
 
 void SAVAGESelectBuffer( ScrnInfoPtr pScrn, int which )
@@ -826,7 +832,6 @@ Bool SAVAGEDRIScreenInit( ScreenPtr pScreen )
     * for canonical symbols in each module.
     */
    if ( !xf86LoaderCheckSymbol( "GlxSetVisualConfigs" ) )	return FALSE;
-   if ( !xf86LoaderCheckSymbol( "DRIScreenInit" ) )		return FALSE;
    if ( !xf86LoaderCheckSymbol( "drmAvailable" ) )		return FALSE;
    if ( !xf86LoaderCheckSymbol( "DRIQueryVersion" ) ) {
       xf86DrvMsg( pScreen->myNum, X_ERROR,
@@ -883,13 +888,16 @@ Bool SAVAGEDRIScreenInit( ScreenPtr pScreen )
    pDRIInfo->ddxDriverMajorVersion = SAVAGE_VERSION_MAJOR;
    pDRIInfo->ddxDriverMinorVersion = SAVAGE_VERSION_MINOR;
    pDRIInfo->ddxDriverPatchVersion = SAVAGE_PATCHLEVEL;
-   pDRIInfo->frameBufferPhysicalAddress = psav->FrameBufferBase;
+
+   pDRIInfo->frameBufferPhysicalAddress = (pointer) psav->FrameBufferBase;
    pDRIInfo->frameBufferSize = psav->videoRambytes;
    pDRIInfo->frameBufferStride = pScrn->displayWidth*(pScrn->bitsPerPixel/8);
    pDRIInfo->ddxDrawableTableEntry = SAVAGE_MAX_DRAWABLES;
 
    /* override default DRI block and wakeup handler */
+   psav->coreBlockHandler = pDRIInfo->wrap.BlockHandler;
    pDRIInfo->wrap.BlockHandler = SAVAGEBlockHandler;
+   psav->coreWakeupHandler = pDRIInfo->wrap.WakeupHandler;
    pDRIInfo->wrap.WakeupHandler = SAVAGEWakeupHandler;
 
    pDRIInfo->wrap.ValidateTree = NULL;
@@ -1530,7 +1538,10 @@ SAVAGEDRIMoveBuffers(WindowPtr pParent, DDXPointRec ptOldOrg,
     }
 
     BCI_SEND(0xc0020000); /* wait for 2D idle */
-    psav->AccelInfoRec->NeedToSync = TRUE;
+    if (psav->useEXA)
+	exaMarkSync(pScreen);
+    else
+	psav->AccelInfoRec->NeedToSync = TRUE;
 }
 
 static void 
@@ -1595,7 +1606,10 @@ SAVAGEDRISubsequentScreenToScreenCopy(
 
 }
 
-
+/*
+ * the FullScreen DRI code is dead, this is just left in place to show how
+ * to set up pageflipping.
+ */
 static Bool
 SAVAGEDRIOpenFullScreen(ScreenPtr pScreen)
 {
